// Copyright 2013 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
/**
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2019 KYOCERA Corporation
 */

package org.chromium.content.browser.picker;

import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.DialogInterface.OnClickListener;
import android.view.KeyEvent;
import android.view.View;

import org.chromium.base.AfpConfigUtil;
import org.chromium.base.AfpUiInterfaceWrapper;
import org.chromium.content.R;
import org.chromium.content.browser.picker.TwoFieldDatePicker.OnMonthOrWeekChangedListener;
import org.chromium.content.browser.picker.TwoFieldDatePicker.SoftkeyStateChangeControl;

/**
 * A two-field data picker dialog.
 */
public abstract class TwoFieldDatePickerDialog extends AlertDialog implements OnClickListener,
        OnMonthOrWeekChangedListener, SoftkeyStateChangeControl {

    private static final String YEAR = "year";
    private static final String POSITION_IN_YEAR = "position_in_year";

    protected final TwoFieldDatePicker mPicker;
    protected final OnValueSetListener mCallBack;

    private int mConfigValueSoftkey;
    private AfpUiInterfaceWrapper mWrapper;

    /**
     * The callback used to indicate the user is done filling in the date.
     */
    public interface OnValueSetListener {

        /**
         * @param year The year that was set.
         * @param positionInYear The position in the year that was set.
         */
        void onValueSet(int year, int positionInYear);
    }

    /**
     * @param context The context the dialog is to run in.
     * @param callBack How the parent is notified that the date is set.
     * @param year The initial year of the dialog.
     * @param weekOfYear The initial week of the dialog.
     */
    public TwoFieldDatePickerDialog(Context context,
             OnValueSetListener callBack,
            int year,
            int positionInYear,
            double minValue,
            double maxValue) {
        this(context, 0, callBack, year, positionInYear, minValue, maxValue);
    }

    /**
     * @param context The context the dialog is to run in.
     * @param theme the theme to apply to this dialog
     * @param callBack How the parent is notified that the date is set.
     * @param year The initial year of the dialog.
     * @param weekOfYear The initial week of the dialog.
     */
    public TwoFieldDatePickerDialog(Context context,
            int theme,
             OnValueSetListener callBack,
            int year,
            int positionInYear,
            double minValue,
            double maxValue) {
        super(context, theme);

        mCallBack = callBack;
        mConfigValueSoftkey = AfpConfigUtil.getInstance().getConfigIntValue(AfpConfigUtil.KEY_PICKER_SOFTKEY);
        mWrapper = new AfpUiInterfaceWrapper(context);

        setIcon(0);

        mPicker = createPicker(context, minValue, maxValue);
        setView(mPicker);
        mPicker.init(year, positionInYear, this);

        mPicker.setSoftkeyStateChangeControl(this);
        setSoftkey(true);
    }

    protected TwoFieldDatePicker createPicker(Context context, double minValue, double maxValue) {
        return null;
    }

    @Override
    public void onClick(DialogInterface dialog, int which) {
        tryNotifyDateSet();
    }

    /**
     * Notifies the listener, if such, that a date has been set.
     */
    protected void tryNotifyDateSet() {
        if (mCallBack != null) {
            mPicker.clearFocus();
            mCallBack.onValueSet(mPicker.getYear(), mPicker.getPositionInYear());
        }
    }

    @Override
    public void onMonthOrWeekChanged(TwoFieldDatePicker view, int year, int positionInYear) {
        mPicker.init(year, positionInYear, null);
    }

    /**
     * Sets the current date.
     *
     * @param year The date week year.
     * @param weekOfYear The date week.
     */
    public void updateDate(int year, int weekOfYear) {
        mPicker.updateDate(year, weekOfYear);
    }

    @Override
    public void onSoftkeyStateChanged(boolean state) {
        setSoftkey(state);
    }

    private void setSoftkey(boolean state) {
        if(mConfigValueSoftkey <= 1) {
            if (state) {
                mWrapper.setDialogSoftkey(this, "webview_softkey_next", AfpUiInterfaceWrapper.LOCAL_RES_ID_NONE, AfpUiInterfaceWrapper.INDEX_CSK);
            } else {
                mWrapper.setDialogSoftkey(this, AfpUiInterfaceWrapper.SET_BRANK_STR, AfpUiInterfaceWrapper.LOCAL_RES_ID_NONE, AfpUiInterfaceWrapper.INDEX_CSK);
            }
        }
    }

    @Override
    public boolean dispatchKeyEvent(KeyEvent event) {
        final int keyCode = event.getKeyCode();
        switch (keyCode) {
            case KeyEvent.KEYCODE_DPAD_CENTER:
            case KeyEvent.KEYCODE_ENTER:
                if (event.getAction() == KeyEvent.ACTION_UP) {
                    if (mConfigValueSoftkey <= 1) {
                        if (mPicker.getLastSoftkeyState()) {
                            if (mPicker.getSoftkeyStateListenSpinner() == mPicker.getYearSpinner()) {
                                mPicker.getPositionInYearSpinner().requestFocus(View.FOCUS_RIGHT);
                            } else {
                                mPicker.getYearSpinner().requestFocus(View.FOCUS_RIGHT);
                            }
                        }
                        return false;
                    } else if (mConfigValueSoftkey == 3){
                        tryNotifyDateSet();
                        dismiss();
                        return true;
                    }
                }
        }
        return super.dispatchKeyEvent(event);
    }
}
